#coding: utf-8
import re
from bs4 import BeautifulSoup
from .generic_utils import str_strip

def str2soup(string):
    """Convert strings to soup, and removed extra tags such as ``<html>``, ``<body>``, and ``<head>``.

    Args:
        string (str) : strings

    Returns:
        bs4.BeautifulSoup : A data structure representing a parsed HTML or XML document.

    Examples:
        >>> string = "<title>Translation-Gummy</title>"
        >>> type(string)
        str
        >>> soup = str2soup(string)
        >>> soup
        <title>Translation-Gummy</title>
        >>> type(soup)
        bs4.BeautifulSoup
        >>> from bs4 import BeautifulSoup
        >>> BeautifulSoup(string)
        <html><head><title>Translation-Gummy</title></head></html>
    """
    soup = BeautifulSoup(markup=string, features="html5lib")
    for attr in ["html", "body", "head"]:
        if hasattr(soup, attr) and getattr(soup, attr) is not None:
            getattr(soup, attr).unwrap()
    return soup
    
def split_section(section, name=None, attrs={}, recursive=True, text=None, **kwargs):
    """ Split ``bs4.BeautifulSoup``.

    Args:
        section (bs4.BeautifulSoup) : A data structure representing a parsed HTML or XML document.
        name (str)                  : A filter on tag name.
        attrs (dict)                : A dictionary of filters on attribute values.
        recursive (bool)            : If this is True, ``.find`` will perform a recursive search of this PageElement's children. Otherwise, only the direct children will be considered.
        text (str)                  : An inner text.
        kwargs (dict)               : A dictionary of filters on attribute values.
    
    Returns:
        list : A list of elements without filter tag elements.

    Examples:
        >>> from bs4 import BeautifulSoup
        >>> from gummy.utils import split_section
        >>> section = BeautifulSoup(\"\"\"
        ... <section>
        ...   <div>
        ...     <h2>Title</h2>
        ...     <div>
        ...     <p>aaaaaaaaaaaaaaaaaaaaaa</p>
        ...     <div>
        ...     <img/>
        ...     </div>
        ...     <p>bbbbbbbbbbbbbbbbbbbbbb</p>
        ...     </div>
        ...   </div>
        ... </section>
        >>> \"\"\")
        >>> len(split_section(section, name="img"))
        3
        >>> split_section(section, name="img")
        [<section>
        <div>
        <h2>Title</h2>
        <div>
        <p>aaaaaaaaaaaaaaaaaaaaaa</p>
        <div>
        </div></div></div></section>,
        <img/>,
        <p>bbbbbbbbbbbbbbbbbbbbbb</p>
        ]
    """
    str_section = str(section)
    page_elements = []
    delimiters = section.find_all(name=name, attrs=attrs, recursive=recursive, text=text, **kwargs)
    # Initialization (Prevent occuring an error when for-loop enter continue at the beginning (i=0))
    end = 0
    for i,delimiter in enumerate(delimiters):
        str_delimiter = str(delimiter)
        start = str_section.find(str_delimiter)
        if start==-1:
            continue
        page_elements.append(str2soup(string=str_section[end:start]))
        page_elements.append(delimiter)
        end = start + len(str_delimiter)
    page_elements.append(str2soup(string=str_section[end:]))
    return page_elements

def group_soup_with_head(soup, name=None, attrs={}, recursive=True, text=None, **kwargs):
    """ Gouping ``bs4.BeautifulSoup`` based on head.

    Args:
        section (bs4.BeautifulSoup) : A data structure representing a parsed HTML or XML document.
        name (str)                  : A filter on tag name.
        attrs (dict)                : A dictionary of filters on attribute values.
        recursive (bool)            : If this is True, ``.find`` will perform a recursive search of this PageElement's children. Otherwise, only the direct children will be considered.
        text (str)                  : An inner text.
        kwargs (dict)               : A dictionary of filters on attribute values.

    Returns:
        list : A list of elements without filter tag elements.

    Examples:
        >>> from bs4 import BeautifulSoup
        >>> from gummy.utils import group_soup_with_head
        >>> section = BeautifulSoup(\"\"\"
        ... <h2>AAA</h2>
        ... <div>
        ...   <p>aaaaaaaaaaaaaaaaaaaaaa</p>
        ... </div>
        ... <h2>BBB</h2>
        ... <div>
        ...   <p>bbbbbbbbbbbbbbbbbbbbbb</p>
        ... </div>
        >>> \"\"\")
        >>> sections = group_soup_with_head(section, name="h2")
        >>> len(sections)
        2
        >>> sections
        [<section><h2>AAA</h2><div>
        <p>aaaaaaaaaaaaaaaaaaaaaa</p>
        </div>
        </section>,
        <section><h2>BBB</h2><div>
        <p>bbbbbbbbbbbbbbbbbbbbbb</p>
        </div>
        </section>]
    """
    str_soup = str(soup)
    sections = []
    heads = soup.find_all(name=name, attrs=attrs, recursive=recursive, text=text, **kwargs)
    # Initialization (Prevent occuring an error when for-loop enter continue at the beginning (i=0))
    end = 0; section = BeautifulSoup(markup="", features="lxml").new_tag(name="section")
    if len(heads)>0:
        for i,head in enumerate(heads):
            str_head = str(head)
            start = str_soup.find(str_head)
            if start==-1:
                continue
            if i>0:
                body = str2soup(string=str_soup[end:start])
                section.append(body)
                sections.append(section)
            end = start + len(str_head)
            section = BeautifulSoup(markup="", features="lxml").new_tag(name="section")
            section.append(head)
        body = str2soup(string=str_soup[end:])
        section.append(body)
        sections.append(section)
    return sections

def replace_soup_tag(soup, 
    new_name, new_namespace=None, new_nsprefix=None, new_attrs={}, new_sourceline=None, 
    new_sourcepos=None, new_kwattrs={},
    old_name=None, old_attrs={}, old_recursive=True, old_text=None, old_limit=None, old_kwargs={}, **kwargs):
    """Replace Old tag with New tag.

    - Args named ``old_XXX`` specifies "How to find old tags"
    - Args named ``new_XXX`` specifies "How to create new tags"

    Args:
        old_name (str)       : A filter on tag name.
        old_attrs (dict)     : A dictionary of filters on attribute values.
        old_recursive (bool) : If this is True, ``.find_all`` will perform a recursive search of this PageElement's children. Otherwise, only the direct children will be considered.
        old_limit (int)      : Stop looking after finding this many results.
        old_kwargs (dict)    : A dictionary of filters on attribute values.
        new_name (str)       : The name of the new Tag.
        new_namespace (str)  : The URI of the new Tag's XML namespace, if any.
        new_prefix (str)     : The prefix for the new Tag's XML namespace, if any.
        new_attrs (dict)     : A dictionary of this Tag's attribute values; can be used instead of `kwattrs` for attributes like 'class' that are reserved words in Python.
        new_sourceline (str) : The line number where this tag was (purportedly) found in its source document.
        new_sourcepos (str)  : The character position within ``sourceline`` where this tag was (purportedly) found.
        new_kwattrs (dict)   : Keyword arguments for the new Tag's attribute values.
    
    Examples:
        >>> from bs4 import BeautifulSoup
        >>> from gummy.utils import replace_soup_tag
        >>> section = BeautifulSoup(\"\"\"
        ... <h2>AAA</h2>
        ... <div>
        ...   <p>aaaaaaaaaaaaaaaaaaaaaa</p>
        ... </div>
        ... <h3>BBB</h3>
        ... <div>
        ...   <p>bbbbbbbbbbbbbbbbbbbbbb</p>
        ... </div>
        >>> \"\"\")
        >>> section = replace_soup_tag(soup=section, old_name="h3", new_name="h2")
        >>> section
        <html><body><h2>AAA</h2>
        <div>
        <p>aaaaaaaaaaaaaaaaaaaaaa</p>
        </div>
        <h2>BBB</h2>
        <div>
        <p>bbbbbbbbbbbbbbbbbbbbbb</p>
        </div>
        </body></html>
    """
    for old in soup.find_all(name=old_name, attrs=old_attrs, recursive=old_recursive, text=old_text, limit=old_limit, **old_kwargs):
        new = BeautifulSoup(markup="", features="lxml").new_tag(name=new_name, namespace=new_namespace, nsprefix=new_nsprefix, attrs=new_attrs, sourceline=new_sourceline, sourcepos=new_sourcepos, **new_kwattrs)
        new.extend(list(old.children))
        old.replace_with(new)
    return soup

def find_target_text(soup, name=None, attrs={}, recursive=True, text=None, default="__NOT_FOUND__", strip=True, **kwargs):
    """Find target element, and get all child strings from it.

    Args:
        soup (bs4.BeautifulSoup) : A data structure representing a parsed HTML or XML document.
        name (str)               : A filter on tag name.
        attrs (dict)             : A dictionary of filters on attribute values.
        recursive (bool)         : If this is True, ``.find`` will perform a recursive search of this PageElement's children. Otherwise, only the direct children will be considered.
        text (str)               : An inner text.
        default (str)            : Default return value if element not found.
        strip (bool)             : Whether to use :func:`str_strip <gummy.utils.generic_utils.str_strip>`
        kwargs (dict)            : A dictionary of filters on attribute values.

    Returns:
        str : text

    Examples:
        >>> from bs4 import BeautifulSoup
        >>> from gummy.utils import find_target_text
        >>> section = BeautifulSoup(\"\"\"
        ... <h2>AAA</h2>
        ... <div> <p>aaaaaaaaaaaaaaaaaaaaaa</p></div>
        >>> \"\"\")
        >>> find_target_text(soup=section, name="div")
        'aaaaaaaaaaaaaaaaaaaaaa'
        >>> find_target_text(soup=section, name="div", strip=False)
        ' aaaaaaaaaaaaaaaaaaaaaa '
        >>> find_target_text(soup=section, name="divdiv", default="not found")
        'not found'
    """
    target = soup.find(name=name, attrs=attrs, recursive=recursive, text=text, **kwargs)
    if target is None:
        text = default
    else:
        text = target.text
    if strip:
        text = str_strip(string=text)
    return text

def find_target_id(soup, key, name=None, attrs={}, recursive=True, text=None, default=None, strip=True, **kwargs):
    """Find target element, and get id from it.

    Args:
        soup (bs4.BeautifulSoup) : A data structure representing a parsed HTML or XML document.
        key (str)                : id name.
        name (str)               : A filter on tag name.
        attrs (dict)             : A dictionary of filters on attribute values.
        recursive (bool)         : If this is True, ``.find`` will perform a recursive search of this PageElement's children. Otherwise, only the direct children will be considered.
        text (str)               : An inner text.
        default (str)            : Default return value if element not found.
        strip (bool)             : Whether to use :func:`str_strip <gummy.utils.generic_utils.str_strip>`
        kwargs (dict)            : A dictionary of filters on attribute values.

    Returns:
        str : text.

    Examples:
        >>> from bs4 import BeautifulSoup
        >>> from gummy.utils import find_target_id
        >>> section = BeautifulSoup(\"\"\"
        ... <h2>IMAGE</h2>
        ... <div>
        ...   <img id="apple-touch-icon" src="https://iwasakishuto.github.io/images/apple-touch-icon/Translation-Gummy.png">
        ... </div>
        >>> \"\"\")
        >>> find_target_id(soup=section, name="img", key="id")
        'apple-touch-icon'
        >>> find_target_id(soup=section, name="img", key="src")
        'https://iwasakishuto.github.io/images/apple-touch-icon/Translation-Gummy.png'
    """
    target = soup.find(name=name, attrs=attrs, recursive=recursive, text=text, **kwargs)
    if target is None:
        id_ = default
    else:
        id_ = target.get(key=key, default=default)
    if strip:
        id_ = str_strip(string=id_)
    return id_


# coding: utf-8
import os
from pathlib import Path

TESTS_DIR_PATH = os.path.dirname(os.path.abspath(__file__))              # path/to/tests
KERASY_LIB_PATH = os.path.join(os.path.dirname(TESTS_DIR_PATH), "gummy") # path/to/kerasy
STOP_FILE = ["__init__.py", "setup.py"] # Files which don't need test.
STOP_DIR = ["templates", "utils"] # Programs under these directories don't need test.

p = Path(KERASY_LIB_PATH)
for abs_path in p.glob("**/*.py"):
    rela_path = abs_path.relative_to(p)
    fn = rela_path.name       # hoge.py
    parent = rela_path.parent # (/path/to)/hoge.py

    # File Name check.
    if fn in STOP_FILE:
        continue

    # Directory name check.
    if any([par.name in STOP_DIR for par in rela_path.parents]):
        continue

    test_prog_path = os.path.join(parent, "test_" + fn)

    if not parent.exists():
        os.makedirs(parent)
        print(f"- [DIR] \033[32m{parent}\033[0m created.")

    if not os.path.exists(test_prog_path):
        with open(test_prog_path, mode="w") as f:
            f.write("# coding: utf-8")
        print(f"\t- [FILE] \033[34m{test_prog_path}\033[0m created.")


# coding: utf-8
""" Utility programs for coloring output. """

__all__ = [
    "toGRAY", "toRED", "toGREEN", "toYELLOW", "toBLUE", "toPURPLE", "toCYAN",
    "toWHITE", "toREVERSE", "toACCENT", "toFLASH", "toRED_FLASH",
]

def _toCOLOR_create(color):
    color = color.upper()
    charcode = {
        "BLACK"     : "",
        "GRAY"      : "\033[30m",
        "RED"       : "\033[31m",
        "GREEN"     : "\033[32m",
        "YELLOW"    : "\033[33m",
        "BLUE"      : "\033[34m",
        "PURPLE"    : "\033[35m",
        "CYAN"      : "\033[36m",
        "WHITE"     : "\033[37m",
        "REVERSE"   : "\033[07m",
        "ACCENT"    : "\033[01m",
        "FLASH"     : "\033[05m",
        "RED_FLASH" : "\033[05;41m",
        "END"       : "\033[0m",
    }.get(color, "\033[34m")
    func = lambda x: f"{charcode}{str(x)}\033[0m"
    func.__doc__ = f"""Convert the output color to {color}
    
    Args:
        x (str): string
        
    Examples:
        >>> from gummy.utils import to{color}
        >>> print(to{color}("hoge"))
        hoge
    """
    return func

toGRAY      = _toCOLOR_create(color="GRAY")
toRED       = _toCOLOR_create(color="RED")
toGREEN     = _toCOLOR_create(color="GREEN")
toYELLOW    = _toCOLOR_create(color="YELLOW")
toBLUE      = _toCOLOR_create(color="BLUE")
toPURPLE    = _toCOLOR_create(color="PURPLE")
toCYAN      = _toCOLOR_create(color="CYAN")
toWHITE     = _toCOLOR_create(color="WHITE")
toREVERSE   = _toCOLOR_create(color="REVERSE")
toACCENT    = _toCOLOR_create(color="ACCENT")
toFLASH     = _toCOLOR_create(color="FLASH")
toRED_FLASH = _toCOLOR_create(color="RED_FLASH")


# coding: utf-8
""" Utility programs for creating HTML or PDF."""
import os
import re
import warnings
import unicodedata
import pdfkit
from jinja2 import Environment, FileSystemLoader

from . import TEMPLATES_DIR
from .generic_utils import str_strip
from .coloring_utils import toRED, toBLUE, toGREEN

def sanitize_filename(fp, dirname=None, ext=None, allow_unicode=False):
    """Convert from original filename to sanitized filename

    Args:
        fp (str)             : File path.
        dirname (str)        : Directory part of the ``fp``
        ext (str)            : Required file extension.
        allow_unicode (bool) : Whether allowing unicode or not.

    Returns:
        str : Sanitized filename.

    Examples:
        >>> from gummy.utils import sanitize_filename
        >>> sanitize_filename("path/to/image\u2013.png")
        >>> 'path/to/image.png'
        >>> sanitize_filename("path/to/image\u2013.jpg", ext=".png")
        >>> 'path/to/image.jpg.png'
        >>> sanitize_filename("path/to/image\u2013.png", allow_unicode=True)
        >>> 'path/to/image–.png'
        >>> # Doesn't work
        >>> sanitize_filename(fp='mir-193 targets ALDH2 and contributes to toxic aldehyde accumulation and tyrosine hydroxylase dysfunction in cerebral ischemia/reperfusion injury')
        'mir-193 targets ALDH2 and contributes to toxic aldehyde accumulation and tyrosine hydroxylase dysfunction in cerebral ischemia/reperfusion injury'
        >>> # Work well :)
        >>> sanitize_filename(fp='mir-193 targets ALDH2 and contributes to toxic aldehyde accumulation and tyrosine hydroxylase dysfunction in cerebral ischemia/reperfusion injury', dirname=".")
        './mir-193 targets ALDH2 and contributes to toxic aldehyde accumulation and tyrosine hydroxylase dysfunction in cerebral ischemia0reperfusion injury'
    """
    if dirname is None:
        dirname, fn = os.path.split(fp)
    else:
        fn = os.path.relpath(fp, start=dirname)
    if allow_unicode:
        fn = unicodedata.normalize("NFKC", fn)
    else:
        fn = unicodedata.normalize("NFKD", fn).encode("ascii", "ignore").decode("ascii")
    fn = str_strip(fn)
    fn = re.sub(pattern=r'[\/\?<>\\:\*\|":]', repl='0', string=fn)
    if ext is not None:
        if not ext.startswith("."): ext = "." + ext
        if not fn.endswith(ext): fn += ext
    fp = os.path.normpath(os.path.join(dirname, fn))
    return fp

def get_jinja_all_attrs(string, keyname):
    """Get the keynames which each element in ``keyname`` is expected to have.

    Args:
        string (str)  : Content strings in template file.
        keyname (str) : keyname which is used in ``template.render(keyname=...)``

    Examples:
        >>> import os
        >>> from gummy.utils import TEMPLATES_DIR, get_jinja_all_attrs
        >>> path = os.path.join(TEMPLATES_DIR, "paper.html")
        >>> with open(path, mode="r", encoding="utf-8") as f:
        ...     html = "".join(f.readlines())
        >>> get_jinja_all_attrs(string=html, keyname="contents")
        {'en', 'head', 'img', 'ja', 'subhead'}
    """
    attributes = set()
    get_from_either_pipe = lambda x,y: x if len(x)>0 else y
    for arg in re.findall(pattern=r"{%\s+for\s+(.+)\s+in\s+" + keyname + r"\s+%}", string=string):
        # get 'bar' from {{ arg.bar }} or {{ arg['bar'] }}
        attrs = re.findall(pattern=rf"{{\s+{arg}(?:\.(.+?)|\[['\"](.+?)['\"]\])\s+}}", string=string)
        attrs = [get_from_either_pipe(*attr) for attr in attrs]
        attributes.update(attrs)
    return attributes

def check_contents(path, contents):
    """ Check whether all attributes in template is contained in contents.

    Args:
        path (str)     : path/to/template.file
        contens (list) : Each element in ``contents`` should be dictionary.
    """
    if (not isinstance(contents, list)) or \
        ((len(contents)>0) and (not isinstance(contents[0], dict))):
        raise TypeError("`contents` should be list, and each element in `contents` should be dictionary.")

    with open(path, mode="r", encoding="utf-8") as f:
        html = "".join(f.readlines())   
    # All attributes in template.
    attributes = get_jinja_all_attrs(string=html, keyname="contents")
    # All keys in contens list.
    content_keys = set([e for content in contents for e in content.keys()])
    
    # Print key which is in content_keys but not in attributes.
    for key in content_keys.difference(attributes):
        warnings.warn(f"An attribute {toGREEN(key)} is not used in {toBLUE(path)}.")
    # Print key which is in attributes but not in content_keys.
    for key in attributes.difference(content_keys):
        warnings.warn(f"An attribute {toGREEN(key)} is not used in this contents, but used in {toBLUE(path)}.")

def tohtml(path, title="", contents=[], searchpath=TEMPLATES_DIR, template="paper.html", verbose=True):
    """ Arrange ``title`` and ``contents`` in html format.

    Args:
        path (str)       : path/to/output.html
        title (str)      : title for html.
        contents (list)  : Contens which used for render method of ``jinja2.environment.Template`` instance.
        searchpath (str) : Loader will find templates from the file system, and this directory is a base.
        template (str)   : template filename. Loader will find ``f"{searchpath}/{template}"``
    
    Returns:
        str : path/to/output.html
    """
    env = Environment(loader=FileSystemLoader(searchpath=searchpath))
    template = env.get_template(template)

    check_contents(path=template.filename, contents=contents)
    
    root,ext = os.path.splitext(path)
    if ext == ".pdf":
        path = root + ".html"
    path = sanitize_filename(fp=path, ext=".html")
    with open(path, mode="w", encoding='utf-8') as f:
        output = template.render(title=title, contents=contents)
        try:
            f.write(output)
        except UnicodeEncodeError:
            f.write(output.encode("utf-8"))

    if verbose: print(f"Save HTML file at {toBLUE(path)}")
    return path

def html2pdf(path, delete_html=True, verbose=True, options={}):
    """Convert from HTML to PDF.

    Args:
        path (str)         : path/to/input.html
        delete_html (bool) : Whether you want to delete html file. (default= ``True``)
        verbose (bool)     : Whether to print message or not. (default= ``True``)
        options (dict)     : options for wkhtmltopdf. See https://wkhtmltopdf.org/usage/wkhtmltopdf.txt
    
    Returns:
        str : path/to/output.pdf
    """
    options.update({
        "page-size"           : "A4",
        "encoding"            : "UTF-8",
        # "quiet"               : not verbose,
        "header-html"         : os.path.join(TEMPLATES_DIR, "header.html"),
        # "include-in-outline"  : True,
        # "load-error-handling" : "ignore",
        # "footer-center"       : "Page  [page]  of  [toPage]",
    })
    html_removed_path = path.replace(".html", "")
    pdf_path = html_removed_path + ".pdf"
    pdfkit.from_file(input=path, output_path=pdf_path, options=options)
    if verbose: print(f"Save PDF file at {toBLUE(pdf_path)}")
    if delete_html:
        os.remove(path)
        if verbose: print(f"Delete original HTML file at {toRED(path)}")
    return pdf_path

def toPDF(path, title="", contents=[], searchpath=TEMPLATES_DIR, template="paper.html", verbose=True, options={}):
    """ Arrange ``title`` and ``contents`` in html format, then convert it to PDF.

    Args:
        path (str)       : path/to/output.html
        title (str)      : title for html.
        contents (list)  : Contens which used for render method of ``jinja2.environment.Template`` instance.
        searchpath (str) : Loader will find templates from the file system, and this directory is a base.
        template (str)   : template filename. Loader will find ``f"{searchpath}/{template}"``
        verbose (bool)   : Whether to print message or not. (default= ``True``)
        options (dict)   : options for wkhtmltopdf. See https://wkhtmltopdf.org/usage/wkhtmltopdf.txt
    
    Returns:
        str : path/to/output.pdf
    """
    pdf_removed_path = path.remove(".pdf", "")
    html_path = pdf_removed_path + ".html"
    html_path = tohtml(path=html_path, title=title, contents=contents, searchpath=searchpath, template=template, verbose=verbose)
    pdf_path = html2pdf(path=html_path, delete_html=True, verbose=verbose, options=options)
    return pdf_path


# coding: utf-8
from gummy import journals
from gummy.utils import get_driver
from gummy.journals import whichJournal
from gummy.utils import check_environ

def _test_gateways(db, gateway, journal_type="nature", **gatewaykwargs):
    url = db.journals.get(journal_type)[0]
    
    # Crawler with gateway.
    gateway_crawler = journals.get(identifier=journal_type, gateway=gateway)
    is_ok, _ = check_environ(
        required_keynames=gateway_crawler.gateway.get_required_keynames(journal_type=journal_type), 
        required_env_varnames=gateway_crawler.gateway.get_required_env_varnames(journal_type=journal_type), 
        verbose=1, 
        **gatewaykwargs,
    )
    if is_ok:
        with get_driver() as driver:
            gateway_title, gateway_contents = gateway_crawler.get_contents(url=url, driver=driver, **gatewaykwargs)

        # Crawler without gateway.
        useless_crawler = journals.get(identifier=journal_type, gateway="useless")
        with get_driver() as driver:
            useless_title, useless_contents = useless_crawler.get_contents(url=url, driver=driver)
            
        assert useless_title == gateway_title
        assert len(useless_contents) < len(gateway_contents)

def test_utokyo_gateways(db):
    _test_gateways(db=db, gateway="utokyo")


# coding: utf-8
import os
import sys
import pytest
from data import TestData
import warnings
try:
    from gummy.utils._warnings import (GummyImprementationWarning, 
                                       EnvVariableNotDefinedWarning)
except ModuleNotFoundError:
    here     = os.path.abspath(os.path.dirname(__file__))
    REPO_DIR = os.path.dirname(here)
    sys.path.append(REPO_DIR)
    print(f"You didn't install 'Translation-Gummy', so add {REPO_DIR} to search path for modules.")
    from gummy.utils._warnings import (GummyImprementationWarning, 
                                       EnvVariableNotDefinedWarning)

def pytest_addoption(parser):
    parser.addoption("--gummy-warnings", choices=["error", "ignore", "always", "default", "module", "once"], default="ignore")

def pytest_configure(config):
    action = config.getoption("gummy_warnings")
    warnings.simplefilter(action, category=GummyImprementationWarning)
    warnings.simplefilter(action, category=EnvVariableNotDefinedWarning)

@pytest.fixture
def db():
    database = TestData()
    return database


# coding: utf-8
""" Utility programs for Selenium WebDriver. See `1. Installation — Selenium Python Bindings 2 documentation <https://selenium-python.readthedocs.io/installation.html#drivers>`_ for more details."""
import time
import warnings
from selenium import webdriver
from selenium.webdriver import DesiredCapabilities
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
# from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC

from ._warnings import DriverNotFoundWarning
from ._path import GUMMY_DIR
from .coloring_utils import toGRAY, toBLUE, toGREEN, toRED
from .generic_utils import get_latest_filename, try_wrapper, handleKeyError

SUPPORTED_DRIVER_TYPES = ["local", "remote"]

def _print_driver_check_log(is_succeed, driver_type):
    """Print logs."""
    if is_succeed:
        state = toGREEN("[success]")
        msg = "driver can be built."
    else:
        state = toRED("[failure]")
        msg = "driver can't be built."
    print(" ".join([state, driver_type, msg]))

def get_chrome_options(browser=False):
    """Get default chrome options.

    Args:
        browser (bool) : Whether you want to run Chrome with GUI browser. (default= ``False`` )

    Examples:
        >>> from gummy.utils import get_chrome_options
        >>> # browser == False (default)
        >>> op = get_chrome_options(browser=False)
        >>> op.arguments
        ['--no-sandbox',
        '--ignore-certificate-errors',
        '--disable-dev-shm-usage',
        '--headless']
        >>> op.experimental_options
        {}
        >>> # browser == True
        >>> op = get_chrome_options(browser=True)
        >>> op.arguments
        ['--no-sandbox',
        '--ignore-certificate-errors',
        '--disable-dev-shm-usage',
        '--kiosk-printing']
        >>> op.experimental_options
        {'prefs': {'profile.default_content_settings.popups': 1,
        'download.default_directory': '/Users/iwasakishuto/.gummy',
        'directory_upgrade': True}}
    """
    chrome_options = Options()
    chrome_options.add_argument('--no-sandbox')
    chrome_options.add_argument('--ignore-certificate-errors')
    chrome_options.add_argument('--disable-dev-shm-usage')
    if not browser:
        chrome_options.add_argument('--headless')
    else:
        chrome_options.add_experimental_option("prefs", {
            # "plugins.always_open_pdf_externally": True,
            "profile.default_content_settings.popups": 1,
            "download.default_directory": GUMMY_DIR,
            "directory_upgrade": True,
        })
        chrome_options.add_argument('--kiosk-printing')

    return chrome_options

def _check_driver(selenium_port="4444"):
    """Check the available drivers. (if one of the drivers is built, there is no problem)"""
    DRIVER_TYPE = ""
    chrome_options = get_chrome_options(browser=False)
    for driver_type in SUPPORTED_DRIVER_TYPES:
        try:
            with eval(f"_get_driver_{driver_type}(chrome_options, selenium_port=selenium_port)") as driver:
                DRIVER_TYPE = driver_type
                _print_driver_check_log(is_succeed=True, driver_type=driver_type)
        except Exception as e:
            _print_driver_check_log(is_succeed=False, driver_type=driver_type)
    return DRIVER_TYPE

def _get_driver_local(chrome_options, **kwargs):
    driver = webdriver.Chrome(options=chrome_options)
    return driver

def _get_driver_remote(chrome_options, **kwargs):
    selenium_port = kwargs.get("selenium_port", "4444")
    driver = webdriver.Remote(
        command_executor=f'http://selenium:{selenium_port}/wd/hub',
        desired_capabilities=DesiredCapabilities.CHROME.copy(),
        options=chrome_options
    )
    return driver

############################
#  START: Check driver
############################

try:
    __DRIVER_SETUP__
except NameError:
    DRIVER_TYPE = _check_driver()
    print(f"DRIVER_TYPE: {toGREEN(DRIVER_TYPE)}")
    __DRIVER_SETUP__ = True
    if DRIVER_TYPE == "":
        warnings.warn(message="Fails to launch all supported drivers.", category=DriverNotFoundWarning)

############################
#  END: Check driver
############################

def get_driver(driver_type=DRIVER_TYPE, chrome_options=None, browser=False, selenium_port="4444"):
    """Get a driver that works in your current environment.

    Args:
        driver_type (str)              : driver type. (default= ``DRIVER_TYPE``)
        chrome_options (ChromeOptions) : Instance of ChromeOptions. If not specify, use :meth:`get_chrome_options() <gummy.utils.driver_utils.get_chrome_options>` to get default options. 
        browser (bool)                 : Whether you want to run Chrome with GUI browser. (default= ``False`` )
        selenium_port (str)            : selenium port number. This will be used when you run on `Docker <https://github.com/iwasakishuto/Translation-Gummy/tree/master/docker>`_
    """
    handleKeyError(lst=SUPPORTED_DRIVER_TYPES, driver_type=driver_type)
    if chrome_options is None: chrome_options = get_chrome_options(browser=browser)
    return eval(f"_get_driver_{driver_type}(chrome_options, selenium_port=selenium_port)")

def try_find_element(driver, by, identifier, timeout=3, verbose=True):
    """Find an element given a By strategy and locator.

    Wrap with :meth:`try_wrapper <gummy.utils.generic_utils.try_wrapper>` 
    to handle the error so that the programs does not stop in a series of processes.

    Args:
        driver (WebDriver) : Selenium WebDriver.
        by (str)           : Locator strategies. See `4. Locating Elements — Selenium Python Bindings 2 documentation <https://selenium-python.readthedocs.io/locating-elements.html>`_
        identifier (str)   : Identifier to find the element
        timeout (int)      : Number of seconds before timing out (default= ``3``)
        verbose (bool)     : Whether you want to print output or not. (default= ``True`` ) 

    Examples:
        >>> from gummy.utils import get_driver, try_find_element
        >>> with get_driver() as driver:
        ...     driver.get("https://www.google.com/")
        ...     e = try_find_element(driver=driver, by="tag name", identifier="img")
        Succeeded to locate element with tag name=img
    """
    return try_wrapper(
        func=WebDriverWait(driver=driver, timeout=timeout).until,
        msg_=f"locate element with {toGREEN(by)}={toBLUE(identifier)}",
        method=lambda x: x.find_element(by=by, value=identifier),
        verbose_=verbose,
    )

def try_find_element_send_keys(driver, by=None, identifier=None, values=(), target=None, timeout=3, verbose=True):
    """Find an element given a By strategy and locator, and Simulates typing into the element.

    Wrap with :meth:`try_wrapper <gummy.utils.generic_utils.try_wrapper>` 
    to handle the error so that the programs does not stop in a series of processes.

    Args:
        driver (WebDriver)  : Selenium WebDriver.
        by (str)            : Locator strategies. See `4. Locating Elements — Selenium Python Bindings 2 documentation <https://selenium-python.readthedocs.io/locating-elements.html>`_
        identifier (str)    : Identifier to find the element
        values (tuple)      : A string for typing, or setting form fields. For setting file inputs, this could be a local file path.
        target (WebElement) : Represents a DOM element. (If you already find element)
        timeout (int)       : Number of seconds before timing out (default= ``3``)
        verbose (bool)      : Whether you want to print output or not. (default= ``True`` ) 
    """
    if target is None:
        target = try_find_element(driver=driver, identifier=identifier, by=by, timeout=timeout)
    if target is not None:
        try_wrapper(
            target.send_keys,
            *tuple(values),
            msg_=f"fill {toBLUE(values)} in element with {toGREEN(by)}={toBLUE(identifier)}",
            verbose_=verbose,
        )
    
def try_find_element_click(driver, by=None, identifier=None, target=None, timeout=3, verbose=True):
    """Find an element given a By strategy and locator, and Clicks the element.

    Wrap with :meth:`try_wrapper <gummy.utils.generic_utils.try_wrapper>` 
    to handle the error so that the programs does not stop in a series of processes.

    Args:
        driver (WebDriver)  : Selenium WebDriver.
        by (str)            : Locator strategies. See `4. Locating Elements — Selenium Python Bindings 2 documentation <https://selenium-python.readthedocs.io/locating-elements.html>`_
        identifier (str)    : Identifier to find the element
        target (WebElement) : Represents a DOM element. (If you already find element)
        timeout (int)       : Number of seconds before timing out (default= ``3``)
        verbose (bool)      : Whether you want to print output or not. (default= ``True`` ) 
    """
    if target is None:
        target = try_find_element(driver=driver, identifier=identifier, by=by, timeout=timeout)
    if target is not None:
        def element_click(driver, target):
            try:
                driver.execute_script("arguments[0].click();", target)
            except StaleElementReferenceException:
                target.click()
        try_wrapper(
            func=element_click,
            msg_=f"click the element with {toGREEN(by)}={toBLUE(identifier)}",
            verbose_=verbose,
            driver=driver,
            target=target,
        )

def click():
    """ function for differentiation """

def pass_forms(driver, **kwargs):
    """Pass through forms.

    You can check the example in :meth:`passthrough_base <gummy.gateways.UTokyoGateWay.passthrough_base>`

    TODO:
        Currently, only ``"id"`` is supported as ``Locator strategies``. Need to 
        be selectable for generality

    Args:
        driver (WebDriver) : Selenium WebDriver.
        kwargs (dict)      : ``key`` is  ``identifier`` and ``val`` is
    """
    for k,v in kwargs.items():
        if callable(v) and v.__qualname__ == "click":
            try_find_element_click(driver=driver, by="id", identifier=k)
        else:
            try_find_element_send_keys(driver=driver, by="id", identifier=k, values=v)

def download_PDF_with_driver(url, dirname=".", verbose=True, timeout=3):
    """Download PDF file with GUI driver.

    Args:
        url (str)       : File URL.
        dirname (str)   : The directory where downloaded data will be saved.
        verbose (bool)  : Whether print verbose or not.
        timeout (int)   : Number of seconds before timing out (default= ``3``)

    Returns:
        path (str) : path/to/downloaded_file
    """
    chrome_options = get_chrome_options(browser=True)
    if "prefs" not in chrome_options._experimental_options:
        chrome_options._experimental_options["prefs"] = {}
    chrome_options._experimental_options["prefs"]["download.default_directory"] = dirname
    chrome_options._experimental_options["prefs"]["plugins.always_open_pdf_externally"] = True
    if verbose: print(f"Downloading PDF from {toBLUE(url)}")
    with get_driver(chrome_options=chrome_options) as driver:
        driver.get(url)
        for _ in range(timeout):
            time.sleep(1)
            path = get_latest_filename(dirname=dirname)
            if not path.endswith(".crdownload"):
                break
    if verbose: print(f"Save PDF at {toBLUE(path)}")
    return path

def wait_until_all_elements(driver, timeout, verbose=True):
    """Wait until all elements visible.
    
    Args:
        driver (WebDriver) : Selenium WebDriver.
        timeout (int)      : Number of seconds before timing out (default= ``3``)
        verbose (bool)     : Whether you want to print output or not. (default= ``True`` ) 
    """
    if verbose: print(f"Wait up to {timeout}[s] for all page elements to load.")
    WebDriverWait(driver=driver, timeout=timeout).until(EC.presence_of_all_elements_located)
    time.sleep(timeout)

def scrollDown(driver, verbose=True):
    """Scroll down to the bottom of the page.
    
    Args:
        driver (WebDriver) : Selenium WebDriver.
        verbose (bool)     : Whether you want to print output or not. (default= ``True`` ) 
    """
    if verbose: print("Scroll down to the bottom of the page.")
    driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")     
    # driver.find_element_by_tag_name('body').click()
    # driver.find_element_by_tag_name('body').send_keys(Keys.PAGE_DOWN)    



# coding: utf-8
import pytest
from gummy.utils import get_driver
from gummy import journals
from data import JournalData

@pytest.mark.parametrize("journal_type", list(JournalData.keys()))
def test_journals(db, journal_type, gateway="useless", **kwargs):
    urls = db.journals.get(journal_type)
    for url in urls:
        crawler = journals.get(identifier=journal_type, gateway=gateway)
        with get_driver() as driver:
            title, texts = crawler.get_contents(url=url, driver=driver)
            
        assert len(title) > 0
        assert len(texts) > 0


# coding: utf-8
import pytest
from gummy.utils import whichJournal, get_driver
from gummy import journals
from data import JournalData

@pytest.mark.parametrize("journal_type", list(JournalData.keys()))
def test_whichJournal(db, journal_type):
    urls = db.journals.get(journal_type)
    with get_driver() as driver:
        for url in urls:
            assert whichJournal(url=url, driver=None)   == journal_type
            assert whichJournal(url=url, driver=driver) == journal_type
            crawler = journals.get(journal_type)
            assert crawler.journal_type == journal_type



# coding: utf-8
""" Utility programs for downloading """
import os
import re
import bs4
import time
import base64
import urllib

from ._path import IMG_NOT_FOUND_SRC
from .coloring_utils  import toBLUE, toGREEN, toRED
from .generic_utils import readable_bytes
from .monitor_utils import progress_reporthook_create
from .driver_utils import download_PDF_with_driver

CONTENT_ENCODING2EXT = {
    "x-gzip"                    : ".gz",
    "image/jpeg"                : ".jpg",
    "image/jpx"                 : ".jpx", 
    "image/png"                 : ".png",
    "image/gif"                 : ".gif",
    "image/webp"                : ".webp",
    "image/x-canon-cr2"         : ".cr2",
    "image/tiff"                : ".tif",
    "image/bmp"                 : ".bmp",
    "image/vnd.ms-photo"        : ".jxr",
    "image/vnd.adobe.photoshop" : ".psd",
    "image/x-icon"              : ".ico",
    "image/heic"                : ".heic",
}

CONTENT_TYPE2EXT = {
    "application/epub+zip"                  : ".epub",
    "application/zip"                       : ".zip",
    "application/x-tar"                     : ".tar",
    "application/x-rar-compressed"          : ".rar",
    "application/gzip"                      : ".gz",
    "application/x-bzip2"                   : ".bz2",
    "application/x-7z-compressed"           : ".7z",
    "application/x-xz"                      : ".xz",
    "application/pdf"                       : ".pdf",
    "application/x-msdownload"              : ".exe",
    "application/x-shockwave-flash"         : ".swf",
    "application/rtf"                       : ".rtf",
    "application/octet-stream"              : ".eot",
    "application/postscript"                : ".ps",
    "application/x-sqlite3"                 : ".sqlite",
    "application/x-nintendo-nes-rom"        : ".nes",
    "application/x-google-chrome-extension" : ".crx",
    "application/vnd.ms-cab-compressed"     : ".cab",
    "application/x-deb"                     : ".deb",
    "application/x-unix-archive"            : ".ar",
    "application/x-compress"                : ".Z",
    "application/x-lzip"                    : ".lz",
}


def decide_extension(content_encoding=None, content_type=None, filename=None):
    """Decide File Extension based on ``content_encoding`` and ``content_type``
    Args:
        content_encoding (str) : The MIME type of the resource or the data.
        content_type (str)     : The Content-Encoding entity header is used to compress the media-type.
        filename (str)         : The filename.
    Returns:
        ext (str): Starts with "."
    Examples:
        >>> from gummy.utils import decide_extension
        >>> decide_extension(content_encoding="x-gzip", content_type="application/zip")
        .gz
        >>> decide_extension(content_encoding="image/png", content_type=None)
        .png
        >>> decide_extension(content_encoding=None, content_type="application/pdf")
        .pdf
    """
    ext = CONTENT_ENCODING2EXT.get(content_encoding) or CONTENT_TYPE2EXT.get(content_type) or "." + str(filename).split(".")[-1]
    return ext

def download_file(url, dirname=".", path=None, bar_width=20, verbose=True):
    """Download a file.
    Args:
        url (str)       : File URL.
        dirname (str)   : The directory where downloaded data will be saved.
        path (str)      : path/to/downloaded_file
        bar_width (int) : The width of progress bar.
        verbose (bool)  : Whether print verbose or not.

    Returns:
        path (str) : path/to/downloaded_file
    
    Examples:
        >>> from gummy.utils import download_file
        >>> download_file(url="https://raw.githubusercontent.com/opencv/opencv/master/data/haarcascades/haarcascade_eye.xml")
        Download a file from https://raw.githubusercontent.com/opencv/opencv/master/data/haarcascades/haarcascade_eye.xml
                    * Content-Encoding : None
                    * Content-Length   : (333.404296875, 'MB')
                    * Content-Type     : text/plain; charset=utf-8
                    * Save Destination : ./haarcascade_eye.xml 
        haarcascade_eye.xml	100.0%[####################] 0.1[s] 5.5[GB/s]	eta -0.0[s]
        './haarcascade_eye.xml'
    """    
    try:
        with urllib.request.urlopen(url) as web_file:
            # Get Information from webfile header
            headers = dict(web_file.headers._headers)
        content_encoding     = headers.get("Content-Encoding")
        content_length, unit = readable_bytes(int(headers.get("Content-Length", 0)))
        content_length       = f"{content_length:.1f} [{unit}]"
        content_type         = headers.get("Content-Type")
        fn = url.split("/")[-1]
        if path is None:
            *name, ext = fn.split(".")
            name = ".".join(name)
            guessed_ext = decide_extension(content_encoding, content_type, fn)
            path = os.path.join(dirname, name+guessed_ext)
        if verbose:
            print(f"""Download a file from {toBLUE(url)}
            * Content-Encoding : {toGREEN(content_encoding)}
            * Content-Length   : {toGREEN(content_length)}
            * Content-Type     : {toGREEN(content_type)}
            * Save Destination : {toBLUE(path)}""")
        _, res = urllib.request.urlretrieve(url=url, filename=path, reporthook=progress_reporthook_create(filename=fn, bar_width=bar_width, verbose=verbose))
    except urllib.error.URLError as e:
        if verbose: 
            print(f"{toRED(e)} : url={toBLUE(url)}")
            print(f"Try to download using webdriver {toRED('(Open Browser)')}")
        try:
            path = download_PDF_with_driver(url=url, dirname=dirname, verbose=verbose)
        except urllib.error.URLError as e:
            if verbose: print(f"{toRED(e)}")
            path = None
    return path

def src2base64(src, base=None):
    """Create base64 encoded img tag from src url or <img> tag element.

    Args:
        src (str, bs4.element.Tag) : Image src url, or ``<img>`` tag element.
        base (str)                 : Base URL. Join a base URL and a possibly relative URL to form an absolute interpretation of the latter.
    
    Returns:
        str : base64 encoded img tag

    Examples:
        >>> from gummy.utils import src2base64
        >>> img_tag = src2base64(src="https://iwasakishuto.github.io/images/apple-touch-icon/Translation-Gummy.png")
        >>> with open("sample.html", mode="w") as f:
        ...     f.write(img_tag)
        >>> # open sample.html to check the results.
        >>> img_tag = src2base64(src="https://iwasakishuto.github.io/images/XXX/XXXXX.png")
        Tried to get an image but got an error: HTTP Error 404: Not Found
        >>> with open("error.html", mode="w") as f:
        ...     f.write(img_tag)
        >>> # open sample.html to check the results.    
    """
    if isinstance(src, bs4.element.Tag) and src.name == "img":
        src = src.get("src") or src.get("data-src") or src.get("data-original")
    # Make URLs which starts with "//" start with "https//"
    # src = re.sub(pattern=r"^(\/\/.*)$", repl=r"https:\1", string=src)
    url = urllib.parse.urljoin(base=base, url=src)
    try:
        request = urllib.request.Request(url, headers={"User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:47.0) Gecko/20100101 Firefox/47.0"}) 
        with urllib.request.urlopen(request) as web_file:
            data = base64.b64encode(web_file.read()).decode('utf-8')
            img_tag = f'<img src="data:image/jpeg;base64,{data}" />'
    except Exception as e:
        print(f"Tried to get an image but got an error: {toRED(e)}")
        img_tag = f'<img src="{IMG_NOT_FOUND_SRC}" />'
    return img_tag
    
def path2base64(path):
    """Create base64 encoded img tag from local image.

    Args:
        path (str) : path/to/image.

    Returns:
        str : base64 encoded img tag

    Examples:
        >>> from gummy.utils import path2base64, download_file
        >>> path = download_file(url="https://iwasakishuto.github.io/images/apple-touch-icon/Translation-Gummy.png")
        Download a file from https://iwasakishuto.github.io/images/apple-touch-icon/Translation-Gummy.png
                    * Content-Encoding : None
                    * Content-Length   : 21.4 [MB]
                    * Content-Type     : image/png
                    * Save Destination : ./Translation-Gummy.png
        Translation-Gummy.png	100.0%[####################] 0.0[s] 3.4[GB/s]	eta -0.0[s]
        >>> img_tag = path2base64(path=path)
        >>> with open("sample.html", mode="w") as f:
        ...     f.write(img_tag)
        >>> # open sample.html to check the results.
    """
    try:
        with open(path, "rb") as image_file:
            data = base64.b64encode(image_file.read()).decode('utf-8')
            img_tag = f'<img src="data:image/jpeg;base64,{data}" />'
    except Exception as e:
        print(toRED(f"[{str(e)}]\nCould not load data from {toBLUE(path)}"))
        img_tag = f'<img src="{IMG_NOT_FOUND_SRC}" />'
    return img_tag



#coding: utf-8
from .coloring_utils  import toBLUE, toGREEN

__all__ = [
    "GummyImprementationWarning", 
    "EnvVariableNotDefinedWarning",
    "DriverNotFoundWarning",
]

class GummyImprementationWarning(Warning):
    """ 
    Warnings that developers will resolve. 
    Developers are now solving in a simple stupid way.
    """

class EnvVariableNotDefinedWarning(Warning):
    """
    Warnings when necessary environment variables are not defined. Use ``write_environ``
    function to define them.

    Examples:

        >>> from gummy.utils import write_environ
        >>> write_environ(
        ...     TRANSLATION_GUMMY_GATEWAY_UTOKYO_USERNAME="username",
        ...     TRANSLATION_GUMMY_GATEWAY_UTOKYO_PASSWORD="password",
        >>> )
    """

class DriverNotFoundWarning(Warning):
    """
    Warnings when launching all supported drivers fails.
    """
    def __init__(self, message):
        super().__init__(message)
        print(
            "Could not create an instance of the Selenium WebDriver. If you want " + \
            "to check the error logs, please call " + toBLUE("gummy.utils.get_driver") + \
            " with specifying the " + toGREEN("driver_type") + " you want to look up." + \
            "If you can not prepare Selenium WebDriver by yourself, please build the environment using Docker." + \
            "Please see " + toBLUE("https://github.com/iwasakishuto/Translation-Gummy/tree/master/docker")
        )
